
(*
  newfileaction is similar to "watch"

  It reads files from a directory,
  and when it finds new files, it starts a program of your choice with the
  filenames as parameters.

  Autor (Urheber/Copyright): Oliver Bandel (Berlin, Germany).

  This is free software: you can redistribute it under the GNU-License GPLv2,
  see file "COPYING" for details.

*)


(** the record-type, that is used for storing the command-line-options *)
type opts_t = { mutable ignoredots: bool; mutable chunk: bool;
                mutable dir: string; mutable progr: string; verbose: bool;
                mutable actinterv: int; mutable loopinterv: int }

(** the command line options *)
let options = { ignoredots = false; chunk = false; dir = "." ; progr = "echo"; verbose = false;
                actinterv = 1; loopinterv = 1
              }

let version = "1.0" (* PROGRAM VERSION *)


(** This module is used for looking up the names in the directory *)
module Nameset = Set.Make(String)

(** This queue holds all new (unprocessed)  filenames *)
let (queue: string Queue.t) = Queue.create ()

(* this mutex is used for the queue-access *)
let queue_lock = Mutex.create ()



(** read the contents of the directory,
      look which wiles are new (not in the already known set)
      and give back this set.  *)
let get_newfiles dir already_known =
  let next_direntry dirh = try Some (Unix.readdir dirh) with End_of_file -> None
  in
    let dh = Unix.opendir dir in
    let rec aux_read set = match next_direntry dh with
        None -> set
      | Some entry -> begin (* add a new file to the set if it matchs all selective criteria *)
                        if
                          ( entry = "." || entry = ".." || (entry.[0] = '.' && options.ignoredots) )
                        then
                          aux_read set (* do not add something; use set as it was *)
                        else
                          aux_read (Nameset.add entry set) (* add entry to set *)
                      end
    in
      let new_files = aux_read Nameset.empty in
    Unix.closedir dh;
    Nameset.diff new_files already_known (* give back set of NEW (not already known) files *)



(** This is an endless-looping function that reads the filenames from the queue
    and starts the program with the filenames as args
*)
let do_action action =
  let names = ref [] in (* here we save all names, that we get from the queue *)

  while true (* endless-loop *)
  do
    if options.verbose then print_endline "do_action!";

    (* get all filenames from the (compilation-unit-global) queue and save in a list *)
    Mutex.lock queue_lock;
      Queue.iter (fun x -> print_endline ("====> " ^ x); names := x :: !names ) queue;
      Queue.clear queue;
    Mutex.unlock queue_lock;

    (* for all filenames call the program *)
    if List.length !names > 0 then
    begin
      if options.chunk
      then
        let cmd = action ^ " " ^ (String.concat " " !names) in
        match Unix.system cmd 
        with
            Unix.WEXITED retval -> ()
          | _ ->  prerr_endline ("action \"" ^ action ^ "\" failed.")
      else
        List.iter (fun name ->
                     match (Unix.system( action ^ " " ^  name )) with
                        Unix.WEXITED retval -> ()
                      | _ ->  prerr_endline ("action \"" ^ action ^ "\" failed.")
                  ) !names;

    end;
    names := []; (* clear list of filenames *)
    Unix.sleep options.actinterv (* sleep for a while *)
  done



(** The mainloop: look for new files and add them into the queue *)
let mainloop dirname =
  let rec aux_mainloop complete_set dirname =
    let newfiles = get_newfiles dirname complete_set in

    if options.verbose then print_endline "* mainloop *";
    (* data critical section! *)
    Mutex.lock queue_lock;
      Nameset.iter (fun x -> Queue.add (dirname ^ "/" ^  x) queue) newfiles; (* add new data to queue *)
    Mutex.unlock queue_lock;

    Unix.sleep options.loopinterv;
    aux_mainloop (Nameset.union complete_set newfiles) dirname
  in
    aux_mainloop Nameset.empty dirname





(** parsing the command line and setting the options-record. *)
let parse () = (*print_endline "OK, starting the parse!";*)
    Arg.parse [
         ("-ai",  Arg.Int (fun intval -> options.actinterv  <- intval ),
             "            action-intervall-time in seconds" );

         ("-li",  Arg.Int (fun intval -> options.loopinterv  <- intval ),
             "            lookup-loop intervall-time in seconds" );

         ("-ignoredots",  Arg.Unit (fun () -> options.ignoredots <- true ),
             "   ignore all dot-files" );

         ("-i",  Arg.Unit (fun () -> options.ignoredots <- true ),
             "            ignore all dot-files" );

         ("-chunk",     Arg.Unit (fun () -> options.chunk    <- true ),
             "        chunk of filenames as parameters (instead individually calling the program for each file)" );
         ("-c",     Arg.Unit (fun () -> options.chunk    <- true ),
             "            chunk of filenames as parameters (instead individually calling the program for each file)" );
         (*
         ("-ias",     Arg.Unit (fun () -> options.ignoreatstart    <- true ), "            ignore all files that will be found at startup)" );
         *)
    
         ("-dir",     Arg.String (fun str -> options.dir    <- str ),
             "          name of the directory (default: \".\")" );
         ("-d",     Arg.String (fun str -> options.dir    <- str ),
             "            name of the directory (default: \".\")" );

         ("-program",     Arg.String (fun str -> options.progr    <- str ),
             "      name of the program to invoke (default: \"echo\")" );

         ("-p",     Arg.String (fun str -> options.progr    <- str ),
             "            name of the program to invoke (default: \"echo\")" );

         ("-v",     Arg.Unit (fun () -> Printf.printf "newfaction: version %s\n" version; exit 0 ),
             "            print program version and exit" );
       (* => DEFAULT (hardcoded) !!!
       *)
              ]
         ( fun name -> prerr_endline "no additional commands accepted"; exit 1  )
         "Use the \"namealign\" tool with following options:"




let _ =
  parse(); (* CLI *)
  ignore(Thread.create do_action options.progr); (* start the action-thread *)
  mainloop options.dir (* mainloop is the loop that looks for new files *)



